import time

import numpy as np

from labml import tracker, logger


# dummy train function
def train():
    return np.random.randint(100)


def main():
    # Reset global step because we incremented in previous loop
    tracker.set_global_step(0)

    for i in range(1, 401):
        tracker.add_global_step()
        loss = train()
        tracker.add(loss=loss)
        if i % 10 == 0:
            tracker.save()
        if i % 100 == 0:
            logger.log()
        time.sleep(0.02)


if __name__ == '__main__':
    main()



import os
import time

from labml import monit
from labml.internal.computer import monitor
from labml.internal.computer.configs import computer_singleton


def is_pid_running(pid):
    import os
    import errno

    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            return False
        elif err.errno == errno.EPERM:
            return True
        else:
            raise
    else:
        return True


def get_running_process():
    pid_file = computer_singleton().config_folder / 'monitor.pid'
    if not pid_file.exists():
        return 0

    with open(str(pid_file), 'r') as f:
        pid = f.read()
        try:
            pid = int(pid)
        except ValueError:
            return 0

        if is_pid_running(pid):
            return pid
        else:
            return 0


def run(is_check_process: bool = True, open_browser: bool = True):
    pid = get_running_process()
    if is_check_process and pid:
        raise RuntimeError(f'This computer is already being monitored. PID: {pid}')

    from uuid import uuid1
    session_uuid = uuid1().hex
    with open(str(computer_singleton().config_folder / 'session.txt'), 'w') as f:
        f.write(session_uuid)

    with open(str(computer_singleton().config_folder / 'monitor.pid'), 'w') as f:
        f.write(str(os.getpid()))

    m = monitor.MonitorComputer(session_uuid, open_browser)

    m.start()

    i = 0
    while True:
        with monit.section('Track', is_new_line=False):
            m.track()
        time.sleep(min(60.0, max(1.0, i / 5.0)))
        i += 1


if __name__ == '__main__':
    run()



import json
from typing import Callable, Any

from labml import lab


def cache(name: str, loader: Callable[[], Any], file_type: str = 'json') -> Any:
    cache_path = lab.get_data_path() / 'cache'
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    path = cache_path / f'{name}.{file_type}'
    if path.exists():
        with open(str(path), 'r') as f:
            if file_type == 'json':
                return json.load(f)
            else:
                raise ValueError(f'Unknown file type: {file_type}')
    else:
        value = loader()
        with open(str(path), 'w') as f:
            if file_type == 'json':
                json.dump(value, f)
            else:
                raise ValueError(f'Unknown file type: {file_type}')

        return value


def cache_get(name: str, file_type: str = 'json') -> Any:
    cache_path = lab.get_data_path() / 'cache'
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    path = cache_path / f'{name}.{file_type}'
    if path.exists():
        with open(str(path), 'r') as f:
            if file_type == 'json':
                return json.load(f)
            else:
                raise ValueError(f'Unknown file type: {file_type}')
    else:
        return None


def cache_set(name: str, value: Any, file_type: str = 'json') -> Any:
    cache_path = lab.get_data_path() / 'cache'
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    path = cache_path / f'{name}.{file_type}'
    with open(str(path), 'w') as f:
        if file_type == 'json':
            json.dump(value, f)
        else:
            raise ValueError(f'Unknown file type: {file_type}')



from typing import Tuple, Union, Callable, Any, List, cast, Sized, TYPE_CHECKING, Iterator

Receiver = Union[str, Callable[[Any], None]]
MyIterator = Union[Sized, int]

if TYPE_CHECKING:
    from ..monitor import Monitor


class SplitIterator:
    def __init__(self, iterator: Tuple[Receiver, MyIterator]):
        receiver, it = iterator
        if isinstance(receiver, str):
            self.name = receiver
            self.callback = None
        else:
            self.name = receiver.__name__
            self.callback = receiver

        if isinstance(it, int):
            self.iterator = range(it)
            self.size = it
        else:
            self.iterator = it
            self.size = len(it)

        self.idx = 0

    def __iter__(self):
        self.idx = 0
        self.it = iter(cast(Iterator, self.iterator))
        return self

    def __len__(self):
        return self.size

    def __next__(self):
        self.idx += 1
        v = next(self.it)
        if self.callback:
            self.callback(v)
            return v
        else:
            return v


class Mix:
    iterators: List[SplitIterator]
    it: List[SplitIterator]

    def __init__(self, *,
                 total_iterations,
                 iterators: List[Tuple[Receiver, MyIterator]],
                 is_monit: bool,
                 logger: 'Monitor'):
        self.logger = logger
        self.is_monit = is_monit
        self.total_iterations = total_iterations
        self.iterators = [SplitIterator(it) for it in iterators]
        self.section = None

    def __iter__(self):
        self.it = [cast(SplitIterator, iter(it)) for it in self.iterators]
        self.idx = 0
        return self

    def __next__(self):
        finished = False
        if self.section is not None:
            self.section.progress(self.current_iterator.idx)
            self.section.__exit__(None, None, None)
            self.section = None
        while not finished:
            finished = True
            for it in self.it:
                if it.idx < len(it):
                    finished = False
                    if it.idx < (self.idx + 1) * len(it) // self.total_iterations:
                        if self.is_monit:
                            self.section = self.logger.section(it.name,
                                                               is_partial=True,
                                                               is_silent=False,
                                                               is_timed=True,
                                                               total_steps=it.size,
                                                               is_children_silent=False,
                                                               is_new_line=True)
                            self.section.__enter__()
                        self.current_iterator = it
                        return it.name, next(it)

            self.idx += 1

        raise StopIteration()


class Enumerate(Sized):
    def __init__(self, it: Sized):
        self.iterator = it

    def __iter__(self):
        self.idx = 0
        self.it = iter(cast(Iterator, self.iterator))
        return self

    def __next__(self):
        idx = self.idx
        self.idx += 1
        return idx, next(self.it)

    def __len__(self) -> int:
        return len(self.iterator)



import threading
import time
from typing import Dict, Optional, TYPE_CHECKING

from labml.internal.api import ApiCaller, Packet, ApiDataSource, ApiResponseHandler
from labml.internal.api.dynamic import DynamicUpdateHandler
from labml.internal.api.url import ApiUrlHandler
from ..configs.processor import ConfigsSaver

if TYPE_CHECKING:
    from ..experiment.experiment_run import Run

LOGS_FREQUENCY = 0


class WebApiConfigsSaver(ConfigsSaver):
    def __init__(self, api_experiment: 'ApiExperiment'):
        self.api_experiment = api_experiment

    def save(self, configs: Dict):
        self.api_experiment.save_configs(configs)


class DynamicHyperParamHandler(ApiResponseHandler):
    def __init__(self, handler: DynamicUpdateHandler):
        self.handler = handler

    def handle(self, data) -> bool:
        if 'dynamic' not in data:
            return False

        self.handler.handle(data['dynamic'])
        return False


class ApiExperiment(ApiDataSource):
    configs_saver: Optional[WebApiConfigsSaver]

    def __init__(self, api_caller: ApiCaller, *,
                 frequency: float,
                 open_browser: bool):
        super().__init__()

        self.frequency = frequency
        self.open_browser = open_browser
        self.api_caller = api_caller
        self.configs_saver = None
        self.data = {}
        self.lock = threading.Lock()

    def get_configs_saver(self):
        if self.configs_saver is None:
            self.configs_saver = WebApiConfigsSaver(self)
        return self.configs_saver

    def save_configs(self, configs: Dict[str, any]):
        with self.lock:
            self.data['configs'] = configs

        self.api_caller.has_data(self)

    def get_data_packet(self) -> Packet:
        with self.lock:
            self.data['time'] = time.time()
            packet = Packet(self.data)
            self.data = {}
            return packet

    def start(self, run: 'Run'):
        self.api_caller.add_handler(ApiUrlHandler(self.open_browser, 'Monitor experiment at '))

        with self.lock:
            self.data.update(dict(
                name=run.name,
                comment=run.comment,
                python_file=run.python_file,
                repo_remotes=run.repo_remotes,
                commit=run.commit,
                commit_message=run.commit_message,
                is_dirty=run.is_dirty,
                start_step=run.start_step,
                load_run=run.load_run,
                tags=run.tags,
                notes=run.notes
            ))

        self.api_caller.has_data(self)

        from labml.internal.api.logs import API_LOGS
        API_LOGS.set_api(self.api_caller, frequency=LOGS_FREQUENCY)

    def set_dynamic_handler(self, handler: DynamicUpdateHandler):
        self.api_caller.add_handler(DynamicHyperParamHandler(handler))

    def status(self, rank: int, status: str, details: str, time_: float):
        with self.lock:
            self.data['status'] = {
                'rank': rank,
                'status': status,
                'details': details,
                'time': time_
            }

        self.api_caller.has_data(self)

        # TODO: Will have to fix this when there are other statuses that dont stop the experiment
        # This will stop the thread after sending all the data
        self.api_caller.stop()



from pathlib import Path
from typing import Optional

from labml import logger, monit
from labml.internal import util
from labml.internal.computer import CONFIGS_FOLDER
from labml.internal.api.configs import WebAPIConfigs
from labml.logger import Text


class Computer:
    """
    ### Computer

    Lab contains the labml specific properties.
    """
    web_api: WebAPIConfigs
    uuid: str
    config_folder: Path

    def __init__(self):
        self.home = Path.home()
        self.config_folder = self.home / CONFIGS_FOLDER
        self.configs_file = self.config_folder / 'configs.yaml'

        self.__load_configs()

    def __load_configs(self):
        if self.config_folder.is_file():
            self.config_folder.unlink()

        if not self.config_folder.exists():
            self.config_folder.mkdir(parents=True)


        if self.configs_file.exists():
            with open(str(self.configs_file)) as f:
                config = util.yaml_load(f.read())
                if config is None:
                    config = {}
        else:
            logger.log([
                ('~/labml/configs.yaml', Text.value),
                ' does not exist. Creating ',
                (str(self.configs_file), Text.meta)])
            config = {}

        if 'uuid' not in config:
            from uuid import uuid1
            config['uuid'] = uuid1().hex
            with open(str(self.configs_file), 'w') as f:
                f.write(util.yaml_dump(config))

        default_config = self.__default_config()
        for k, v in default_config.items():
            if k not in config:
                config[k] = v

        self.uuid = config['uuid']
        web_api_url = config['web_api']
        if web_api_url[0:4] != 'http':
            web_api_url = f"https://api.labml.ai/api/v1/computer?labml_token={web_api_url}&"
        self.web_api = WebAPIConfigs(url=web_api_url,
                                     frequency=config['web_api_frequency'],
                                     verify_connection=config['web_api_verify_connection'],
                                     open_browser=config['web_api_open_browser'],
                                     is_default=web_api_url == self.__default_config()['web_api'])

    def set_token(self, token: str):
        with monit.section('Update ~/labml/configs.yaml'):
            with open(str(self.configs_file)) as f:
                config = util.yaml_load(f.read())
                assert config is not None

            config['web_api'] = token

            with open(str(self.configs_file), 'w') as f:
                f.write(util.yaml_dump(config))

    def __str__(self):
        return f"<Computer uuid={self.uuid}>"

    def __repr__(self):
        return str(self)

    @staticmethod
    def __default_config():
        return dict(
            web_api='https://api.labml.ai/api/v1/computer?',
            web_api_frequency=0,
            web_api_verify_connection=True,
            web_api_open_browser=True,
        )


_internal: Optional[Computer] = None


def computer_singleton() -> Computer:
    global _internal
    if _internal is None:
        _internal = Computer()

    return _internal



from typing import Dict, List

import psutil


class ProcessInfo:
    key: int
    pid: int
    name: str

    def __init__(self, key: int, pid: int, name: str):
        self.key = key
        self.pid = pid
        self.name = name

        self.alive = True
        self.active = True


class ProcessMonitor:
    pids: Dict[int, int]
    processes: List[ProcessInfo]

    def __init__(self):
        self.pids = {}
        self.processes = []
        self.data = {}

    def track_process(self, p: psutil.Process):
        with p.oneshot():
            key = None
            if p.pid in self.pids:
                key = self.pids[p.pid]
                if self.processes[key].name != p.name():
                    key = None
            if key is None:
                key = len(self.processes)
                self.processes.append(ProcessInfo(key,
                                               p.pid,
                                               p.name()))
                self.pids[p.pid] = key
                self.data.update({
                    f'process.{key}.name': p.name(),
                    f'process.{key}.pid': p.pid,
                })

            self.processes[key].active = True

            try:
                res = p.memory_info()
                self.data.update({
                    f'process.{key}.rss': res.rss,
                    f'process.{key}.vms': res.vms,
                })
            except psutil.AccessDenied:
                pass
            try:
                res = p.memory_percent()
                self.data.update({
                    f'process.{key}.mem': res,
                })
            except psutil.AccessDenied:
                pass

            try:
                res = p.cpu_percent()
                self.data.update({
                    f'process.{key}.cpu': res,
                })
            except psutil.AccessDenied:
                pass

            try:
                res = p.num_threads()
                self.data.update({
                    f'process.{key}.threads': res,
                })
            except psutil.AccessDenied:
                pass

    def track(self, data):
        self.data = data

        for p in self.processes:
            p.active = False

        for p in psutil.process_iter():
            self.track_process(p)

        for p in self.processes:
            if not p.active and p.alive:
                p.alive = False
                self.data.update({
                    f'process.{p.key}.dead': True,
                })




import os
import sys
from typing import Optional

from labml.logger import Text

from labml import logger, monit


class Service:
    def __init__(self):
        from labml.internal.computer.configs import computer_singleton
        self.home = computer_singleton().home
        self.service_path = computer_singleton().home / '.config' / 'systemd' / 'user' / 'labml.service'

    def _get_service_file_content(self):
        argv = sys.argv
        # assert len(argv) == 2
        # assert argv[1] == 'monitor'
        lines = [
            '[Unit]',
            'Description=labml.ai monitoring service',
            'After=network.target',
            '',
            '[Service]',
            'Type=simple',
            f'ExecStart={argv[0]} service-run',
            'Restart=always',
            '',
            '[Install]',
            'WantedBy=default.target',
        ]

        return '\n'.join(lines)

    def _create(self):
        if self.service_path.exists():
            logger.log(f'{self.service_path} exists')
            return

        if not self.service_path.parent.exists():
            self.service_path.parent.mkdir(parents=True)

        with open(str(self.service_path), 'w') as f:
            f.write(self._get_service_file_content())

        with monit.section('Reload systemd daemon'):
            ret = os.system('systemctl --user daemon-reload')
            if ret != 0:
                monit.fail()

        with monit.section('Enable service'):
            ret = os.system('systemctl --user enable labml.service')
            if ret != 0:
                monit.fail()

    def create(self):
        self._create()

        with monit.section(f'Create {self.service_path}'):
            logger.log(['Start the monitoring service with:\n',
                        ('systemctl --user start labml.service', Text.value),
                        '\n, stop with:\n',
                        ('systemctl --user stop labml.service', Text.value),
                        '\n and check status with:\n',
                        ('systemctl --user status labml.service', Text.value),
                        ])

    def set_token(self):
        from labml.internal.computer.configs import computer_singleton

        if not computer_singleton().web_api.is_default:
            return True

        while True:
            token = input('Enter app.labml.ai token (Go to Settings after logging into app.labml.ai):')

            if len(token) != 32:
                logger.log("Invalid token", Text.danger)

            break

        computer_singleton().set_token(token)


_internal: Optional[Service] = None


def service_singleton() -> Service:
    global _internal
    if _internal is None:
        _internal = Service()

    return _internal


def _test():
    s = Service()
    print(s)
    # s.create()
    s.set_token()


if __name__ == '__main__':
    _test()



import argparse
import subprocess
import threading
import time
import webbrowser
from typing import List

from typing.io import IO

from labml import logger, experiment
from labml.experiment import generate_uuid
from labml.internal.api import ApiCaller, SimpleApiDataSource
from labml.internal.api.logs import ApiLogs
from labml.logger import Text


def _open_dashboard():
    try:
        import labml_dashboard
    except (ImportError, ModuleNotFoundError):
        logger.log("Cannot import ", ('labml_dashboard', Text.highlight), '.')
        logger.log('Install with ',
                   ('pip install labml_dashboard', Text.value))
        return

    labml_dashboard.start_server()


class ExecutorThread(threading.Thread):
    process: subprocess.Popen

    def __init__(self, command: str, api_logs: ApiLogs):
        super().__init__(daemon=False)
        self.api_logs = api_logs
        self.command = command
        self.exit_code = 0

    def _read(self, stream: IO, name: str):
        while stream.readable():
            data = stream.read(1024).decode('utf-8')
            if len(data) == 0:
                break
            print(data)
            self.api_logs.outputs(**{name: data})

    def _read_stdout(self):
        self._read(self.process.stdout, 'stdout_')

    def _read_stderr(self):
        self._read(self.process.stderr, 'stderr_')

    def run(self):
        self.process = subprocess.Popen(self.command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        while True:
            self._read_stdout()
            self._read_stderr()
            if self.process.poll() is not None:
                break
            time.sleep(0.1)
        self.exit_code = self.process.returncode

        self._read_stdout()
        self._read_stderr()


def _capture(args: List[str]):
    api_caller = ApiCaller("https://api.lab-ml.com/api/v1/track?", {'run_uuid': generate_uuid()})
    api_logs = ApiLogs()
    data = {
        'name': 'Capture',
        'comment': ' '.join(args),
        'time': time.time()
    }

    def _started(url):
        if url is None:
            return None

        logger.log([('Monitor experiment at ', Text.meta), (url, Text.link)])
        webbrowser.open(url)

    api_caller.has_data(SimpleApiDataSource(data, callback=_started))
    api_logs.set_api(api_caller, frequency=0)

    thread = ExecutorThread(' '.join(args), api_logs)
    thread.start()
    thread.join()
    data = {
        'rank': 0,
        'status': 'completed',
        'details': None,
        'time': time.time()
    }

    api_caller.has_data(SimpleApiDataSource({
        'status': data,
        'time': time.time()
    }))

    api_caller.stop()


def _launch(args: List[str]):
    import sys
    import os

    if 'RUN_UUID' not in os.environ:
        os.environ['RUN_UUID'] = experiment.generate_uuid()

    cwd = os.getcwd()
    if 'PYTHONPATH' in os.environ:
        python_path = os.environ['PYTHONPATH']
        print(python_path)
        os.environ['PYTHONPATH'] = f"{python_path}:{cwd}:{cwd}/src"
    else:
        os.environ['PYTHONPATH'] = f"{cwd}:{cwd}/src"

    cmd = [sys.executable, '-u', '-m', 'torch.distributed.launch', *args]
    print(cmd)
    try:
        process = subprocess.Popen(cmd, env=os.environ)
        process.wait()
    except Exception as e:
        logger.log('Error starting launcher', Text.danger)
        raise e

    if process.returncode != 0:
        logger.log('Launcher failed', Text.danger)
        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd)


def _monitor():
    from labml.internal.computer import process
    from labml.internal.computer.configs import computer_singleton

    process.run(True, computer_singleton().web_api.open_browser)


def _service():
    from labml.internal.computer.service import service_singleton

    service_singleton().set_token()
    service_singleton().create()


def _service_run():
    from labml.internal.computer import process

    process.run(True, False)


def main():
    parser = argparse.ArgumentParser(description='labml.ai CLI')
    parser.add_argument('command', choices=['dashboard', 'capture', 'launch', 'monitor', 'service', 'service-run'])
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if args.command == 'dashboard':
        _open_dashboard()
    elif args.command == 'capture':
        _capture(args.args)
    elif args.command == 'launch':
        _launch(args.args)
    elif args.command == 'monitor':
        _monitor()
    elif args.command == 'service':
        _service()
    elif args.command == 'service-run':
        _service_run()
    else:
        raise ValueError('Unknown command', args.command)



import sys
import threading
import time
from io import StringIO
from typing import Optional

from labml.internal.api import ApiCaller, ApiDataSource, Packet

WARMUP_COMMITS = 5


class ApiLogs(ApiDataSource):
    api_caller: Optional[ApiCaller]
    frequency: float

    def __init__(self):
        super().__init__()

        self.api_caller = None
        self.frequency = 1
        self.last_committed = time.time()
        self.commits_count = 0
        self.data = {}
        self.lock = threading.Lock()

    def set_api(self, api_caller: ApiCaller, *, frequency: float):
        self.api_caller = api_caller
        self.frequency = frequency
        self.check_and_flush()

    def check_and_flush(self):
        if self.api_caller is None:
            return
        with self.lock:
            if not self.data:
                return
        t = time.time()
        freq = self.frequency
        if self.commits_count < WARMUP_COMMITS:
            freq /= 2 ** (WARMUP_COMMITS - self.commits_count)
        if self.data.get('stderr', '') != '' or self.commits_count == 0 or t - self.last_committed > freq:
            self.commits_count += 1
            self.api_caller.has_data(self)

    def get_data_packet(self) -> Packet:
        with self.lock:
            self.last_committed = time.time()
            self.data['time'] = time.time()
            packet = Packet(self.data)
            self.data = {}
            return packet

    def outputs(self, *,
                stdout_: str = '',
                stderr_: str = '',
                logger_: str = ''):
        with self.lock:
            if stdout_ != '':
                self.data['stdout'] = self.data.get('stdout', '') + stdout_
            if stderr_ != '':
                self.data['stderr'] = self.data.get('stderr', '') + stderr_
            if logger_ != '':
                self.data['logger'] = self.data.get('logger', '') + logger_

        self.check_and_flush()


API_LOGS = ApiLogs()


class OutputStream(StringIO):
    def write(self, *args, **kwargs):  # real signature unknown
        super().write(*args, **kwargs)
        save = StringIO()
        save.write(*args, **kwargs)
        API_LOGS.outputs(**{self.type_: save.getvalue()})
        self.original.write(*args, **kwargs)

    def __init__(self, original, type_):  # real signature unknown
        super().__init__()
        self.type_ = type_
        self.original = original


_original_stdout_write = sys.stdout.write
_original_stderr_write = sys.stderr.write


def _write_stdout(*args, **kwargs):
    _original_stdout_write(*args, **kwargs)
    save = StringIO()
    save.write(*args, **kwargs)
    API_LOGS.outputs(stdout_=save.getvalue())


def _write_stderr(*args, **kwargs):
    _original_stderr_write(*args, **kwargs)
    save = StringIO()
    save.write(*args, **kwargs)
    API_LOGS.outputs(stderr_=save.getvalue())


def capture():
    sys.stdout.write = _write_stdout
    sys.stderr.write = _write_stderr


capture()



import json
import socket
import threading
import time
import urllib.error
import urllib.request
from dataclasses import dataclass
from queue import Queue
from typing import Dict, Optional, Set, List

import labml
from labml import logger
from labml.logger import Text
from labml.utils.notice import labml_notice

UPDATING_APP_MESSAGE = 'Updating App. Please wait'


@dataclass
class Packet:
    data: Dict[str, any]
    idx: int = -1


class ApiDataSource:
    def get_data_packet(self) -> Packet:
        raise NotImplementedError


class SimpleApiDataSource(ApiDataSource):
    def __init__(self, data: Dict[str, any]):
        self.data = data

    def get_data_packet(self) -> Packet:
        return Packet(data=self.data)


class ApiResponseHandler:
    def handle(self, data) -> bool:
        raise NotImplementedError


class _WebApiThread(threading.Thread):
    def __init__(self, url: str, *, timeout_seconds: int, daemon: bool):
        super().__init__(daemon=daemon)
        self.please_wait_count = 0
        self.timeout_seconds = timeout_seconds
        self.url = url
        self.queue: Queue[ApiDataSource] = Queue()
        self.is_stopped = False
        self.errored = False
        self.handlers: List[ApiResponseHandler] = []

    def push_data_source(self, data_source: ApiDataSource):
        self.queue.put(data_source)

    def stop(self):
        self.is_stopped = True
        logger.log('Still updating app.labml.ai, please wait for it to complete...', Text.highlight)
        self.please_wait_count = 1

    def add_handler(self, handler: ApiResponseHandler):
        self.handlers.append(handler)

    @staticmethod
    def _is_updating_notification(packet: Packet):
        """
        Checks if we just printed app updating notification.
        If so filter those out.
        """
        data = packet.data
        if set(data.keys()) != {'stdout', 'time'} and set(data.keys()) != {'logger', 'time'}:
            return False

        if 'stdout' in data:
            lines = data['stdout'].split('\n')
        else:
            lines = data['logger'].split('\n')

        if len(lines) == 1 and lines[0].find(UPDATING_APP_MESSAGE) != -1:
            return True

        return False

    def _get_packets(self) -> List[Packet]:
        sources = [self.queue.get()]
        while not self.queue.empty():
            sources.append(self.queue.get())

        sources = list(set(sources))

        packets = [s.get_data_packet() for s in sources]
        return [p for p in packets if not self._is_updating_notification(p)]

    def run(self):
        while True:
            packets = self._get_packets()
            if self.is_stopped:
                if not packets:
                    logger.log()
                    logger.log('Finished updating LabML App.', Text.highlight)
                    return

                logger.log(UPDATING_APP_MESSAGE + '...' * self.please_wait_count, Text.meta,
                           is_new_line=False)
                self.please_wait_count += 1
            retries = 0
            while True:
                if self._process(packets):
                    break
                else:
                    logger.log(f'Retrying again in 10 seconds ({retries})...', Text.highlight)
                    time.sleep(10)
                    retries += 1
            if self.is_stopped and self.queue.empty():
                time.sleep(0.5)
                if self.queue.empty():
                    logger.log()
                    logger.log('Finished updating LabML App.', Text.highlight)
                    return

    def _process(self, packets: List[Packet]) -> bool:
        if not packets:
            return True

        data = [p.data for p in packets]

        try:
            response = self._send(data)
        except urllib.error.HTTPError as e:
            labml_notice([f'Failed to send to {self.url}: ',
                          (str(e.code), Text.value),
                          '\n' + str(e.reason)])
            return False
        except urllib.error.URLError as e:
            labml_notice([f'Failed to connect to {self.url}\n',
                          str(e.reason)])
            return False
        except socket.timeout as e:
            labml_notice([f'{self.url} timeout\n',
                          str(e)])
            return False
        except ConnectionResetError as e:
            labml_notice([f'Connection reset by LabML App server {self.url}\n',
                          str(e)])
            return False

        for h in self.handlers:
            if h.handle(response):
                break

        return True

    def _send(self, data: List[Dict[str, any]]) -> Dict:
        req = urllib.request.Request(self.url)
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        data_json = json.dumps(data)
        data_json = data_json.encode('utf-8')
        req.add_header('Content-Length', str(len(data)))

        response = urllib.request.urlopen(req, data_json, timeout=self.timeout_seconds)
        content = response.read().decode('utf-8')
        result = json.loads(content)

        for e in result.get('errors', []):
            if 'error' in e:
                labml_notice(['LabML App Error: ', (e['error'] + '', Text.key), '\n',
                              (e['message'], Text.value)],
                             is_danger=True)
                self.errored = True
                raise RuntimeError('LabML App Error', e)
            elif 'warning' in e:
                labml_notice(
                    ['LabML App Warning: ', (e['warning'] + ': ', Text.key), (e['message'], Text.value)])
            else:
                self.errored = True
                raise RuntimeError('Unknown error from LabML App', e)

        return result


class ApiCaller:
    web_api_url: str
    params: Dict[str, str]
    thread: Optional[_WebApiThread]
    state_attributes: Set[str]

    def __init__(self, web_api_url: str, params: Dict[str, str], *,
                 timeout_seconds: int = 15,
                 daemon: bool = False):
        super().__init__()

        self.daemon = daemon
        params.copy()
        params['labml_version'] = labml.__version__
        params = '&'.join([f'{k}={v}' for k, v in params.items()])

        self.web_api_url = f'{web_api_url}{params}'
        self.timeout_seconds = timeout_seconds
        self.thread = None
        self.stopped = False

    def _check(self) -> bool:
        if self.stopped:
            return False

        if self.thread is None:
            self.thread = _WebApiThread(self.web_api_url,
                                        timeout_seconds=self.timeout_seconds,
                                        daemon=self.daemon)

        if self.thread.errored:
            raise RuntimeError('LabML App error: See above for error details')

        return True

    def has_data(self, source: ApiDataSource):
        if not self._check():
            return False

        self.thread.push_data_source(source)
        if not self.thread.is_alive():
            self.thread.start()

    def add_handler(self, handler: ApiResponseHandler):
        if not self._check():
            return False

        self.thread.add_handler(handler)

    def stop(self):
        self.stopped = True
        self.thread.stop()



import warnings
from typing import Tuple, Union

from labml.logger import Text

from labml import tracker, logger


class DynamicHyperParam:
    def __init__(self, default: Union[float, int], type_: str, range_: Tuple[float, float]):
        self._type_ = type_
        self._range_ = range_
        self._default = default
        self._value = default
        self._key = None
        self._change_tracked = False
        self._last_tracked_step = -1

    def __call__(self):
        if not self._change_tracked or self._last_tracked_step != tracker.get_global_step():
            if self._key is None:
                warnings.warn('Register dynamic schedules with `experiment.configs` to update them live from the app')
            else:
                tracker.add(f'hp.{self._key}', self._value)

        self._change_tracked = True
        self._last_tracked_step = tracker.get_global_step()

        return self._value

    def register(self, key):
        self._key = key

    def set_value(self, value):
        if self._key is None:
            warnings.warn('Register dynamic schedules with `experiment.configs` to update them live from the app')
        else:
            logger.log([
                ('Dynamic hyper parameter changed: ', Text.subtle),
                (self._key, Text.key),
                (': ', Text.subtle),
                (f'{self._value}', Text.value),
                (' -> ', Text.subtle),
                (f'{value}', Text.value),
            ])
        self._change_tracked = False
        self._value = value

    def to_yaml(self):
        return {
            'type': 'DynamicSchedule',
            'default': self._default,
            'range': list(self._range_),
            'dynamic_type': self._type_
        }


class FloatDynamicHyperParam(DynamicHyperParam):
    def __init__(self, default: float, range_: Tuple[float, float] = (0, 1)):
        super().__init__(default, 'float', range_)


class IntDynamicHyperParam(DynamicHyperParam):
    def __init__(self, default: int, range_: Tuple[int, int] = (1, 16)):
        super().__init__(default, 'int', range_)

    def set_value(self, value):
        super().set_value(int(value))



import threading
import time
from typing import Dict
from typing import Union

import numpy as np
from labml.internal.api.url import ApiUrlHandler

from labml.internal.api import ApiCaller, Packet, ApiDataSource

MAX_BUFFER_SIZE = 1024

LOGS_FREQUENCY = 0


class Header(ApiDataSource):
    def __init__(self, api_caller: ApiCaller, *,
                 frequency: float,
                 open_browser: bool):
        super().__init__()

        self.frequency = frequency
        self.open_browser = open_browser
        self.api_caller = api_caller
        self.name = None
        self.comment = None
        self.data = {}
        self.lock = threading.Lock()

    def get_data_packet(self) -> Packet:
        with self.lock:
            self.data['time'] = time.time()
            packet = Packet(self.data)
            self.data = {}
            return packet

    def start(self, configs: Dict[str, any]):
        self.api_caller.add_handler(ApiUrlHandler(self.open_browser, 'Monitor computer at '))
        with self.lock:
            self.data['configs'] = configs
            self.data['name'] = 'My computer'

        self.api_caller.has_data(self)

    def status(self, rank: int, status: str, details: str, time_: float):
        with self.lock:
            self.data['status'] = {
                'rank': rank,
                'status': status,
                'details': details,
                'time': time_
            }

        self.api_caller.has_data(self)

        self.api_caller.stop()


class Writer(ApiDataSource):
    def __init__(self, api_caller: ApiCaller, *,
                 frequency: float):
        super().__init__()

        self.frequency = frequency
        self.api_caller = api_caller
        self.last_committed = time.time()
        self.data = {}
        self.lock = threading.Lock()

    @staticmethod
    def _parse_key(key: str):
        return key

    def track(self, indicators: Dict[str, Union[float, int]]):
        t = time.time()
        for k, v in indicators.items():
            if k not in self.data:
                self.data[k] = []
            self.data[k].append((t, v))

        t = time.time()
        freq = self.frequency
        if t - self.last_committed > freq:
            self.flush()

    def get_data_packet(self) -> Packet:
        with self.lock:
            self.last_committed = time.time()
            self.data['track'] = self.get_and_clear_indicators()
            self.data['time'] = time.time()
            packet = Packet(self.data)
            self.data = {}
            return packet

    def flush(self):
        with self.lock:
            if not self.data:
                return

        self.api_caller.has_data(self)

    def get_and_clear_indicators(self):
        data = {}

        for key, value in self.data.items():
            value = np.array(value)
            timestamp: np.ndarray = value[:, 0]
            value: np.ndarray = value[:, 1]
            while value.shape[0] > MAX_BUFFER_SIZE:
                if value.shape[0] % 2 == 1:
                    value = np.concatenate((value, value[-1:]))
                    timestamp = np.concatenate((timestamp, timestamp[-1:]))

                n = value.shape[0] // 2
                timestamp = np.mean(timestamp.reshape(n, 2), axis=-1)
                value = np.mean(value.reshape(n, 2), axis=-1)

            data[key] = {
                'step': timestamp.tolist(),
                'value': value.tolist()
            }

        self.data = {}

        return data



from pathlib import Path

from labml import monit


def download_file(url: str, path: Path):
    if path.exists():
        return

    import urllib.request

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    with monit.section("Download"):
        def reporthook(count, block_size, total_size):
            monit.progress(count * block_size / total_size)

        urllib.request.urlretrieve(url, path, reporthook=reporthook)



from argparse import Namespace
from typing import Any, Dict, Optional, Union

import torch
from pytorch_lightning.loggers.base import LightningLoggerBase, rank_zero_experiment
from pytorch_lightning.utilities import rank_zero_only

from labml import experiment, tracker, lab
from labml.internal.experiment import experiment_singleton


class LabMLLightningLogger(LightningLoggerBase):
    def __init__(self):
        super().__init__()

    @property
    @rank_zero_experiment
    def experiment(self):
        return None

    @rank_zero_only
    def log_hyperparams(self, params: Union[Dict[str, Any], Namespace]) -> None:
        params = self._convert_params(params)
        params = self._flatten_dict(params)
        experiment.configs(params)

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, Union[torch.Tensor, float]], step: Optional[int] = None) -> None:
        if step is None:
            tracker.add_global_step()
            tracker.save(metrics)
        else:
            tracker.save(step, metrics)

    def reset_experiment(self):
        pass

    @property
    def save_dir(self) -> Optional[str]:
        return str(lab.get_experiments_path())

    @property
    def name(self) -> str:
        return str(experiment_singleton().run.name)

    @property
    def version(self) -> str:
        return experiment_singleton().run.uuid



import time

from labml import tracker, experiment


def main():
    experiment.create(name='Test')

    with experiment.start():
        for i in range(1, 401):
            tracker.add_global_step()
            time.sleep(1)
            tracker.save(loss=1.)


if __name__ == '__main__':
    main()



class WebAPIConfigs:
    url: str
    frequency: float
    verify_connection: bool
    open_browser: bool

    def __init__(self, *,
                 url: str,
                 frequency: float,
                 verify_connection: bool,
                 open_browser: bool,
                 is_default: bool):
        self.is_default = is_default
        self.open_browser = open_browser
        self.frequency = frequency
        self.verify_connection = verify_connection
        self.url = url



import webbrowser

from labml import logger
from labml.internal.api import ApiResponseHandler
from labml.logger import Text


class ApiUrlHandler(ApiResponseHandler):
    def __init__(self, open_browser: bool, label: str):
        self.label = label
        self.open_browser = open_browser
        self.notified = False

    def handle(self, data) -> bool:
        if self.notified:
            return False

        if 'url' not in data:
            return False

        url = data['url']
        logger.log([(self.label, Text.meta), (url, Text.link)])
        if self.open_browser:
            webbrowser.open(url)

        self.notified = True

        return False



from fastai.callback.core import Callback, to_detach, patch
from fastai.learner import Learner

from labml import tracker, experiment


class LabMLFastAICallback(Callback):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def before_fit(self):
        pass

    def after_batch(self):
        tracker.add_global_step()
        if self.training:
            metrics = {'loss.train': self.learn.loss}
        else:
            metrics = {'loss.valid': self.learn.loss}
        try:
            for m in self.learn.metrics:
                if m.value is not None:
                    metrics[m.name] = m.value
        except:
            pass

        tracker.save(metrics)

    def after_epoch(self):
        metrics = {}
        try:
            for m in self.learn.metrics:
                if m.value is not None:
                    metrics[m.name] = m.value
        except:
            pass

        tracker.save(metrics)
        tracker.new_line()

    def after_fit(self):
        pass


@patch
def labml_configs(self: Learner):
    configs = {}
    try:
        configs['n_epoch'] = self.learn.n_epoch
        configs['model_class'] = str(type(self.learn.model))
    except:
        pass

    return configs



from typing import List

from labml.internal.logger.destinations import Destination
from labml.internal.util import is_ipynb, is_ipynb_pycharm


def create_destination() -> List[Destination]:
    from labml.internal.logger.destinations.console import ConsoleDestination

    if is_ipynb():
        if is_ipynb_pycharm():
            from labml.internal.logger.destinations.ipynb_pycharm import IpynbPyCharmDestination
            return [IpynbPyCharmDestination(), ConsoleDestination(False)]
        else:
            from labml.internal.logger.destinations.ipynb import IpynbDestination
            return [IpynbDestination(), ConsoleDestination(False)]
    else:
        return [ConsoleDestination(True)]



import time
from pathlib import Path

import psutil

from labml.internal.api import ApiCaller
from labml.internal.computer.configs import computer_singleton
from labml.internal.computer.monitor.process import ProcessMonitor
from labml.internal.computer.writer import Writer, Header
from labml.utils.notice import labml_notice


class MonitorComputer:
    def __init__(self, session_uuid: str, open_browser):
        api_caller = ApiCaller(computer_singleton().web_api.url,
                               {'computer_uuid': computer_singleton().uuid, 'session_uuid': session_uuid},
                               timeout_seconds=15,
                               daemon=True)
        self.writer = Writer(api_caller, frequency=computer_singleton().web_api.frequency)
        self.header = Header(api_caller,
                             frequency=computer_singleton().web_api.frequency,
                             open_browser=open_browser)
        self.data = {}
        self.cache = {}
        self.nvml = None
        self.n_gpu = 0

        self.process_monitor = ProcessMonitor()

        try:
            from py3nvml import py3nvml as nvml
            self.nvml = nvml
        except ImportError:
            labml_notice('Install py3nvml to monitor GPUs:\n pip install py3nvml',
                         is_warn=False)

    def start(self):
        configs = {
            'os': self.get_os(),
            'cpu.logical': psutil.cpu_count(),
            'cpu.physical': psutil.cpu_count(logical=False),
        }

        configs.update(self.get_gpu_header())

        self.header.start(configs)

        self.first()

    def track_gpu(self):
        if not self.nvml:
            return

        self.nvml.nvmlInit()
        for i in range(self.n_gpu):
            handle = self.nvml.nvmlDeviceGetHandleByIndex(i)
            self.data.update({
                f'gpu.memory.used.{i}': self.nvml.nvmlDeviceGetMemoryInfo(handle).used,
                f'gpu.utilization.{i}': self.nvml.nvmlDeviceGetUtilizationRates(handle).gpu,
                f'gpu.temperature.{i}': self.nvml.nvmlDeviceGetTemperature(handle, self.nvml.NVML_TEMPERATURE_GPU),
                f'gpu.power.usage.{i}': self.nvml.nvmlDeviceGetPowerUsage(handle),
            })

        self.nvml.nvmlShutdown()

    def first_gpu(self):
        if not self.nvml:
            return

        self.nvml.nvmlInit()
        for i in range(self.n_gpu):
            handle = self.nvml.nvmlDeviceGetHandleByIndex(i)
            self.data.update({
                f'gpu.memory.total.{i}': self.nvml.nvmlDeviceGetMemoryInfo(handle).total,
                f'gpu.power.limit.{i}': self.nvml.nvmlDeviceGetPowerManagementLimit(handle),
            })

        self.nvml.nvmlShutdown()

    def get_gpu_header(self):
        if not self.nvml:
            return {}

        self.nvml.nvmlInit()
        self.n_gpu = self.nvml.nvmlDeviceGetCount()
        res = {'gpus': self.n_gpu}
        for i in range(self.n_gpu):
            handle = self.nvml.nvmlDeviceGetHandleByIndex(i)
            res.update({
                f'gpu.name.{i}': self.nvml.nvmlDeviceGetName(handle),
            })
        self.nvml.nvmlShutdown()

        return res

    def track_net_io_counters(self):
        res = psutil.net_io_counters()
        t = time.time()
        if 'net.recv' in self.cache:
            td = t - self.cache['net.time']
            self.data.update({
                'net.recv': (res.bytes_recv - self.cache['net.recv']) / td,
                'net.sent': (res.bytes_sent - self.cache['net.sent']) / td,
            })
        self.cache['net.recv'] = res.bytes_recv
        self.cache['net.sent'] = res.bytes_sent
        self.cache['net.time'] = t

    def track_memory(self):
        res = psutil.virtual_memory()
        self.data.update({
            'memory.total': res.total,
            'memory.used': res.used,
            'memory.available': res.available,
        })

    def track_cpu(self):
        res = psutil.cpu_times()
        self.data.update({
            'cpu.idle': res.idle,
            'cpu.system': res.system,
            'cpu.user': res.user,
        })
        res = psutil.cpu_freq()
        if res is not None:
            self.data.update({
                'cpu.freq': res.current,
                'cpu.freq.min': res.min,
                'cpu.freq.max': res.max,
            })
        res = psutil.cpu_percent(percpu=True)
        self.data.update({f'cpu.perc.{i}': p for i, p in enumerate(res)})

    def track_disk(self):
        res = psutil.disk_usage(Path.home())
        self.data.update({
            'disk.free': res.free,
            'disk.total': res.total,
            'disk.used': res.used,
        })

    def track_processes(self):
        self.process_monitor.track(self.data)

    def track(self):
        self.track_net_io_counters()
        # inspect(psutil.net_if_addrs())
        # inspect(psutil.net_if_stats())
        self.track_memory()
        self.track_cpu()
        self.track_disk()
        self.track_gpu()
        # self.track_processes()

        self.writer.track(self.data)
        self.data = {}

    def first(self):
        # self.track_memory()
        # self.track_cpu()
        # self.track_disk()
        self.first_gpu()

        self.writer.track(self.data)
        self.data = {}

    @staticmethod
    def get_os():
        if psutil.MACOS:
            return 'macos'
        elif psutil.LINUX:
            return 'linux'
        elif psutil.WINDOWS:
            return 'windows'
        else:
            return 'unknown'



import typing
from typing import Optional, List, Union, Tuple

from labml.internal.util.colors import StyleCode
from .iterator import Iterator
from .loop import Loop
from .mix import Mix
from .sections import Section, OuterSection
from ..logger import logger_singleton as logger
from ..logger.types import LogPart
from ..tracker import tracker_singleton as tracker
from ...logger import Text
from ...utils.notice import labml_notice


class Monitor:
    __loop_indicators: List[Union[str, Tuple[str, Optional[StyleCode]]]]
    __is_looping: bool

    def __init__(self):
        self.__loop: Optional[Loop] = None
        self.__sections: List[Section] = []
        self.__is_looping = False
        self.__loop_indicators = []
        self.__is_silent = False

    def clear(self):
        self.__loop: Optional[Loop] = None
        self.__sections: List[Section] = []
        self.__is_looping = False
        self.__loop_indicators = []

    def silent(self, is_silent: bool = True):
        self.__is_silent = is_silent

    def mix(self, total_iterations, iterators: List[Tuple[str, typing.Sized]],
            is_monit: bool):
        return Mix(total_iterations=total_iterations,
                   iterators=iterators,
                   is_monit=is_monit, logger=self)

    def iterate(self, name, iterable: Union[typing.Iterable, typing.Sized, int],
                total_steps: Optional[int], *,
                is_silent: bool,
                is_children_silent: bool,
                is_timed: bool,
                section: Optional[Section]):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=total_steps,
                        is_children_silent=is_children_silent,
                        is_enumerate=False,
                        section=section)

    def enum(self, name, iterable: typing.Sized, *,
             is_silent: bool,
             is_children_silent: bool,
             is_timed: bool,
             section: Optional[Section]):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=None,
                        is_children_silent=is_children_silent,
                        is_enumerate=True,
                        section=section)

    def section(self, name, *,
                is_silent: bool,
                is_timed: bool,
                is_partial: bool,
                is_new_line: bool,
                is_children_silent: bool,
                total_steps: float) -> Section:
        if self.__is_looping:
            if len(self.__sections) != 0:
                is_silent = True

            section = self.__loop.get_section(name=name,
                                              is_silent=is_silent,
                                              is_timed=is_timed,
                                              is_partial=is_partial,
                                              total_steps=total_steps,
                                              parents=[s.name for s in self.__sections])
            self.__sections.append(section)
        else:
            if len(self.__sections) > 0:
                if self.__sections[-1].is_silent or self.__sections[-1].is_children_silent:
                    is_silent = True
                    is_children_silent = True
            self.__sections.append(OuterSection(monitor=self,
                                                name=name,
                                                is_silent=is_silent,
                                                is_timed=is_timed,
                                                is_partial=is_partial,
                                                is_new_line=is_new_line,
                                                is_children_silent=is_children_silent,
                                                total_steps=total_steps,
                                                level=len(self.__sections)))

        return self.__sections[-1]

    def progress(self, steps: float):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report progress")

        if self.__sections[-1].progress(steps):
            self.__log_line()

    def set_successful(self, is_successful=True):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report success")

        self.__sections[-1].is_successful = is_successful
        self.__log_line()

    def loop(self, iterator_: typing.Collection, *,
             is_track: bool,
             is_print_iteration_time: bool):
        if len(self.__sections) != 0:
            labml_notice(['LabML Loop: ', ('Starting loop inside sections', Text.key), '\n',
                          (
                              'This could be because some iterators crashed in a previous cell in a notebook.',
                              Text.meta)],
                         is_danger=False)
            err = RuntimeError('Section outside loop')
            for s in reversed(self.__sections):
                s.__exit__(type(err), err, err.__traceback__)
            # raise RuntimeError("Cannot start a loop within a section")

        self.__loop = Loop(iterator=iterator_,
                           monitor=self,
                           is_track=is_track,
                           is_print_iteration_time=is_print_iteration_time)
        return self.__loop

    def start_loop(self):
        self.__is_looping = True
        tracker().start_loop(self.set_looping_indicators)

    def finish_loop(self):
        if len(self.__sections) != 0:
            raise RuntimeError("Cannot be within a section when finishing the loop")
        tracker().finish_loop()
        self.__loop = None
        self.__is_looping = False

    def section_enter(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Entering a section without creating a section.\n"
                               "Always use logger.section to create a section")

        if section is not self.__sections[-1]:
            raise RuntimeError("Entering a section other than the one last_created\n"
                               "Always user with logger.section(...):")

        if len(self.__sections) > 1 and not self.__sections[-2].is_parented:
            self.__sections[-2].make_parent()
            if not self.__is_silent and not self.__sections[-1].is_silent:
                logger().log([])

        self.__log_line()

    def __log_looping_line(self):
        parts = [(f"{tracker().global_step :8,}:  ", Text.highlight)]
        parts += self.__loop.log_sections()
        parts += self.__loop_indicators
        parts += self.__loop.log_progress()

        if not self.__is_silent:
            logger().log(parts, is_new_line=False)

    def __log_line(self):
        if self.__is_looping:
            self.__log_looping_line()
            return

        if len(self.__sections) == 0:
            return

        parts = self.__sections[-1].log()
        if parts is None:
            return

        if not self.__is_silent:
            logger().log(parts, is_new_line=False)

    def set_looping_indicators(self, indicators: List[LogPart]):
        self.__loop_indicators = indicators
        self.__log_looping_line()

    def section_exit(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Impossible")

        if section is not self.__sections[-1]:
            raise RuntimeError("Impossible")

        self.__log_line()
        self.__sections.pop(-1)


_internal: Optional[Monitor] = None


def monitor_singleton() -> Monitor:
    global _internal
    if _internal is None:
        _internal = Monitor()

    return _internal
